from weml_validator.attribute_validators import AttributeRuleRequired, AttributeRuleEnum, AttributeRuleOptional, \
    AttributeRuleMaxLength
from weml_validator.repository import ValidatorRepository
from weml_validator.tag_validators import EmptyTagValidator, ChildrenSubsetValidator, CombinedValidator

validator_instance = ValidatorRepository.get_instance()

INLINES = ["", "a", "br", "w-entity", "w-format", "w-lang", "w-non-egw", "w-note", "w-page", "w-sent"]
BLOCKS = ["figure", "w-list", "w-text-block", "table", "hr"]
CONTAINERS = ["w-heading", "w-page", "w-para", "w-para-group"]

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='div',
    allowed_children=CONTAINERS,
    expected_child_count=1
))

# region Container elements
validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-heading",
    attribute_rules={
        "skip": [AttributeRuleEnum(None, "1")],
        "level": [AttributeRuleRequired(), AttributeRuleEnum("1", "2", "3", "4", "5", "6")]
    },
    allowed_children=["w-text-block"],
    required_children={"w-text-block": 1},
    expected_child_count=1
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-para",
    attribute_rules={
        "skip": [AttributeRuleEnum(None, "1")],
        "indent": [
            AttributeRuleEnum(None, "0", "-1", "-2", "-3", "-4", "-5", "-6", "-7", "-8", "-9", "-10", "1", "2", "3",
                              "4", "5", "6", "7", "8", "9", "10")],
        "align": [AttributeRuleEnum(None, "left", "right", "center", "justify")],
        "role": [AttributeRuleEnum(None, "address", "addressee", "author", "date", "place", "introduction",
                                   "letterhead", "salutation", "signature-date", "signature-line", "signature-source",
                                   "bible-text", "devotional-text", "poem-source", "publication-info", "title")]
    },
    allowed_children=["w-text-block"],
    required_children={"w-text-block": 1},
    expected_child_count=1
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-para-group",
    attribute_rules={
        "skip": [AttributeRuleEnum(None, "1")],
    },
    allowed_children=["w-para"],
    required_children={"w-para": None}
))

# endregion

# region Block elements
validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-text-block",
    attribute_rules={
        "type": [AttributeRuleEnum("paragraph", "blockquote", "poem", None)],
    },
    allowed_children=INLINES
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='figure',
    allowed_children=["img", "figcaption"],
    required_children={"img": 1, "figcaption": lambda x: x <= 1},
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='figcaption',
    allowed_children=["w-text-block"],
    required_children={"w-text-block": 1},
))

validator_instance.add_validator(EmptyTagValidator(
    tag='img',
    attribute_rules={
        "src": [AttributeRuleRequired()],
        "alt": [AttributeRuleOptional()]
    }
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='w-list',
    attribute_rules={
        "type": [AttributeRuleEnum(None, "ordered", "unordered")]
    },
    allowed_children=["w-li"],
    required_children={"w-li": None}
))
validator_instance.add_validator(ChildrenSubsetValidator(
    tag='w-li',
    attribute_rules={
    },
    allowed_children=BLOCKS,
    expected_child_count=1
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='table',
    allowed_children=["thead", "tbody"],
    required_children={"thead": lambda x: x <= 1, "tbody": lambda x: x <= 1},
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='thead',
    allowed_children=["tr"],
))
validator_instance.add_validator(ChildrenSubsetValidator(
    tag='tbody',
    allowed_children=["tr"],
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='tr',
    allowed_children=["th", "td"],
))
table_cell_rules = {
    "align": [AttributeRuleEnum(None, "left", "right", "center")],
    "valign": [AttributeRuleEnum(None, "top", "middle", "bottom")],
    "colspan": [AttributeRuleOptional(),
                AttributeRuleEnum(None, "1", "2", "3", "4", "5", "6", "7", "8", "9", "10")],
    "rowspan": [AttributeRuleOptional(),
                AttributeRuleEnum(None, "1", "2", "3", "4", "5", "6", "7", "8", "9", "10")],
}
validator_instance.add_validator(ChildrenSubsetValidator(
    tag='th',
    attribute_rules=table_cell_rules,
    allowed_children=BLOCKS,
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag='td',
    attribute_rules=table_cell_rules,
    allowed_children=BLOCKS,
))

validator_instance.add_validator(EmptyTagValidator(
    tag='hr'
))

# endregion

# region Inlines
validator_instance.add_validator(CombinedValidator(
    "a",
    ChildrenSubsetValidator(
        tag="a",
        allowed_children=INLINES,
        attribute_rules={
            "href": [AttributeRuleRequired()]
        },
        unique=True
    ),
    EmptyTagValidator(
        tag="a",
        attribute_rules={
            "id": [AttributeRuleRequired()]
        }
    )
))

validator_instance.add_validator(EmptyTagValidator(
    tag="br",
    attribute_rules={}
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-entity",
    attribute_rules={
        "type": [
            AttributeRuleRequired(),
            AttributeRuleEnum("addressee", "date", "place", "topic", "topic-word")
        ],
        "value": [AttributeRuleOptional()]
    },
    allowed_children=INLINES
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-format",
    attribute_rules={
        "type": [
            AttributeRuleRequired(),
            AttributeRuleEnum("bold", "italic", "underline", "superscript", "subscript", "small-caps",
                              "all-caps")
        ],
    },
    allowed_children=INLINES
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-lang",
    attribute_rules={
        "lang": [AttributeRuleRequired(), AttributeRuleMaxLength(5)],
        "dir": [AttributeRuleEnum(None, "ltr", "rtl")]
    },
    allowed_children=INLINES
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-non-egw",
    attribute_rules={
        "type": [
            AttributeRuleRequired(),
            AttributeRuleEnum("appendix", "comment", "foreword", "intro", "note", "preface", "text")
        ],
    },
    allowed_children=INLINES))
validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-note",
    attribute_rules={
        "type": [
            AttributeRuleEnum(None, "footnote", "endnote")
        ]
    },
    allowed_children=["w-note-header", "w-note-body"],
    required_children={"w-note-header": 1, "w-note-body": 1},
    expected_child_count=2
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-note-body",
    allowed_children=["w-text-block"],
    required_children={"w-text-block": None}

))
validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-note-header",
    allowed_children=[""],
    required_children={"": None}
))

validator_instance.add_validator(EmptyTagValidator(
    tag="w-page",
    attribute_rules={
        "number": [AttributeRuleRequired()]
    }
))

validator_instance.add_validator(ChildrenSubsetValidator(
    tag="w-sent",
    allowed_children=INLINES,
    unique=True
))

# endregion

__all__ = ["validator_instance"]



import bs4

from weml_validator import ValidationResult, ValidationError
from weml_validator.attribute_validators import AttributeRule


class ValidatorBase:
    def __init__(self, tag, attribute_rules: dict[str, list[AttributeRule]]):
        self._tag = tag
        self._attribute_rules = attribute_rules

    @property
    def tag(self):
        return self._tag

    def validate(self, tag: bs4.Tag) -> ValidationResult:
        result = ValidationResult.success()
        if tag.name != self.tag:  # pragma: no cover
            result.add_node_error("Invalid tag", tag)
        for attr, rule_set in self._attribute_rules.items():
            attr_value = tag.attrs.get(attr, None)
            for rule in rule_set:
                if not rule.validate(attr_value):
                    result.add_node_error(f"Attribute {attr} failed validation with rule: {rule}", tag)
        return result


class ValidatorRepository:
    def __init__(self):
        self._validators = {}

    def add_validator(self, validator: ValidatorBase):
        self._validators[validator.tag] = validator

    def get_validator(self, tag: str) -> ValidatorBase:
        return self._validators[tag]

    def validate(self, element: bs4.Tag, base_tag) -> ValidationResult:
        tag = element.name
        if tag not in self._validators:
            return ValidationResult(False, [ValidationError(tag, 0, 0)])
        validator = self.get_validator(tag)
        return validator.validate(element)

    @classmethod
    def get_instance(cls):
        return _instance


_instance = ValidatorRepository()

__all__ = ["ValidatorRepository", "ValidatorBase"]



import bs4

from weml_validator.errors import ValidationResult, ValidationError
from weml_validator.validators import validator_instance


def validate_weml_document(document: str) -> ValidationResult:  # pragma: no cover
    return ValidationResult(False, [])


def validate_weml_paragraph(node_html: str, base_tag: str | None = None) -> ValidationResult:
    div_content = f"<div>\n{node_html}\n</div>"
    validation_result = validate_weml_element(div_content, base_tag)
    for error in validation_result.errors:
        error.line -= 1
    return validation_result


def validate_weml_element(node_html: str, base_tag: str | None = None) -> ValidationResult:
    content = bs4.BeautifulSoup(node_html, "html.parser")
    for node in content:
        result = validator_instance.validate(node, base_tag)
        if not result:
            return result
    return ValidationResult(True, [])


__all__ = ["validate_weml_document", "validate_weml_paragraph", "validate_weml_element",
           "ValidationResult", "ValidationError"]



from typing import Callable

import bs4

from weml_validator import ValidationResult
from weml_validator.attribute_validators import AttributeRule
from weml_validator.repository import ValidatorRepository, ValidatorBase


class EmptyTagValidator(ValidatorBase):
    """Validator for tags that must be empty"""

    def __init__(self, tag, attribute_rules: dict[str, list[AttributeRule]] = None):
        """Create a new EmptyTagValidator
        :param tag: The tag name
        :param attribute_rules: A dictionary of attribute rules. The key is the attribute name, the value is a list of
            AttributeRule objects
        """
        super().__init__(tag, attribute_rules or {})

    def validate(self, tag: bs4.Tag) -> ValidationResult:
        result = super().validate(tag)
        if tag.contents:
            result.add_node_error("Tag must be empty", tag)
        return result


class CombinedValidator(ValidatorBase):
    """Validator that combines multiple validators. One of the validators must pass for the tag to be valid"""

    def __init__(self, tag: str, *validators: ValidatorBase):
        super().__init__(tag, {})
        self._validators = validators

    def validate(self, tag: bs4.Tag) -> ValidationResult:
        output_result = ValidationResult.success()
        for validator in self._validators:
            result = validator.validate(tag)
            if result.is_valid:
                return ValidationResult.success()
            output_result = output_result.combine_with(result)
        return output_result


class ChildrenSubsetValidator(ValidatorBase):

    def __init__(self, tag,
                 allowed_children: list[str],
                 attribute_rules: dict[str, list[AttributeRule]] | None = None,
                 required_children: dict[str, int | None | Callable[[int], bool]] | None = None,
                 unique=False,
                 expected_child_count: int | None = None):
        """ Create a new ChildrenSubsetValidator
        :param tag: The tag name
        :param attribute_rules: A dictionary of attribute rules. The key is the attribute name, the value is a list of
            AttributeRule objects
        :param allowed_children: A list of allowed children tag names. Empty string represents text nodes
        :param required_children: A dictionary of required children. The key is the child tag name, the value is the
            number of required children. If the value is None, at least one child is required
        :param unique: If true, this tag cannot have any parent with the same tag name
        """
        super().__init__(tag, attribute_rules or {})
        self._allowed_children = allowed_children
        self._required_children = required_children or {}
        self._is_unique = unique
        self._expected_child_count = expected_child_count

    def validate(self, tag: bs4.Tag) -> ValidationResult:
        result = super().validate(tag)
        children = tag.children

        if self._is_unique:
            parents = tag.find_parents()
            if any(parent.name == tag.name for parent in parents):
                result.add_node_error(f"Tag must be unique", tag)
        children_count = {}
        for child in children:
            if isinstance(child, bs4.NavigableString):
                child_name = ""
                if child.strip() != "":
                    if "" not in self._allowed_children:
                        result.add_node_error(f"Text inside the tag is not allowed", tag)
            elif isinstance(child, bs4.Tag):
                child_name = child.name
                if child_name not in self._allowed_children:
                    result.add_node_error(f"Invalid child `{child_name}` (allowed {self._allowed_children})", tag)
                result = result.combine_with(ValidatorRepository.get_instance().validate(child, tag.name))
            else:  # pragma: no cover
                # todo: How to test this?
                continue
            children_count[child_name] = children_count.get(child_name, 0) + 1
        for child_name, expected_count in self._required_children.items():
            if expected_count is None:
                if child_name not in children_count:
                    result.add_node_error(f"Required child `{child_name}` missing", tag)

            elif isinstance(expected_count, int):
                if children_count.get(child_name, 0) != expected_count:
                    result.add_node_error(f"Required child `{child_name}` missing", tag)
            elif callable(expected_count):
                if not expected_count(children_count.get(child_name, 0)):
                    result.add_node_error(f"Required child `{child_name}` missing", tag)
        if self._expected_child_count is not None:
            total_children = 0
            for count in (val for c, val in children_count.items() if c):
                total_children += count
            if total_children != self._expected_child_count:
                result.add_node_error(
                    f"Invalid number of children. Got {total_children}, expected: {self._expected_child_count}: {children_count}",
                    tag
                )

        return result



import bs4


class ValidationError:
    def __init__(self, message: str, line: int, column: int):
        self.message = message
        self.line = line
        self.column = column

    def __repr__(self): # pragma: no cover
        return f"Error at {self.line}:{self.column}: {self.message}"


class ValidationResult:
    def __init__(self, is_valid: bool, errors: list[ValidationError]):
        self._is_valid = is_valid
        self._errors = errors

    def __bool__(self):
        return self._is_valid

    @property
    def is_valid(self):
        return self._is_valid

    @property
    def errors(self):
        return self._errors

    @staticmethod
    def success():
        return ValidationResult(True, [])

    def add_node_error(self, message: str, node: bs4.Tag):
        self._errors.append(ValidationError(f"Error in {node}: {message}", node.sourceline, node.sourcepos))
        self._is_valid = False

    def __iadd__(self, other):
        is_valid = self.is_valid and other.is_valid
        errors = self.errors + other.errors
        return ValidationResult(is_valid, errors)

    def combine_with(self, other: 'ValidationResult'):
        return self.__iadd__(other)



import abc
from abc import ABC


class AttributeRule(ABC):
    @abc.abstractmethod
    def validate(self, value: str | None):
        """ do nothing"""


class AttributeRuleRequired(AttributeRule):
    def validate(self, value: str | None):
        return value is not None and value != ""


class AttributeRuleOptional(AttributeRule):
    def validate(self, value: str | None):
        return True


class AttributeRuleEnum(AttributeRule):
    def __init__(self, *allowed_values: str | None):
        self._allowed_values = allowed_values

    def validate(self, value: str | None):
        return value in self._allowed_values


class AttributeRuleMaxLength(AttributeRule):

    def __init__(self, max_length: int):
        super().__init__()
        self._max_length = max_length

    def validate(self, value: str | None):
        return value is None or len(value) <= self._max_length


__all__ = ["AttributeRule", "AttributeRuleRequired", "AttributeRuleOptional", "AttributeRuleEnum",
           "AttributeRuleMaxLength"]
